#!/usr/bin/env python3

from asyncio import Future
from collections import defaultdict, deque
from typing import Any, Dict, List, Tuple, Deque

from fastapi import FastAPI, HTTPException, Request

from . import core

waitings: Dict[
    str, Deque[Tuple[Future, Request]]
] = defaultdict(deque)

app: FastAPI = FastAPI()

JSONDict = Dict[str, Any]


async def notice(token: str) -> None:
    lst = waitings[token]
    while lst:
        future, request = lst.popleft()
        if not await request.is_disconnected():
            future.set_result("ok")
            break
        future.set_result("disconnected")


@app.post("/new/{state}", status_code=201)
async def _(state: str, data: JSONDict = {}) -> int:
    id = core.new(state, data)
    await notice(state)
    return id


@app.post("/lock/{state}")
async def _(state: str, request: Request, wait: bool = False) -> JSONDict:
    while True:
        try:
            return core.lock(state)
        except core.NotAllowed as e:
            raise HTTPException(400, e.args[0])
        except core.NotFound:
            if not wait:
                raise HTTPException(404, f"{state} is not found")
            future = Future()
            waitings[state].append((future, request))
            if await future == "disconnected":
                return {}  # this request has gone, return any has no meaning
            continue


@app.post("/transit/{id}/{state}", status_code=204)
async def _(id: int, state: str, data: JSONDict = {}) -> None:
    try:
        core.transit(id, state, data)
        await notice(state)
    except core.NotFound:
        raise HTTPException(404)
    except core.NotAllowed as e:
        raise HTTPException(403, " -> ".join(e.args))


@app.get("/list-locked")
def _() -> List[JSONDict]:
    return core.list_locked()


@app.get("/list/{state}")
def _(state: str) -> List[JSONDict]:
    return core.list_available(state)


@app.get("/{id}")
def _(id: int) -> JSONDict:
    try:
        return core.info(id)
    except core.NotFound:
        raise HTTPException(404)


def main():
    pass


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

from datetime import datetime, timedelta
from typing import Any, Dict, List, Union
from os import getenv

from pony import orm

from .entities import Fsm, db


prefix_locked = getenv("FSMHUB_PREFIX_LOCKED") or "."
ageing_time = timedelta(seconds=float(getenv("FSMHUB_AGEING_SECONDS") or 300))

JSONDict = Dict[str, Any]


class NotFound(Warning):
    pass


class NotAllowed(Warning):
    pass


def new(state: str, data: JSONDict = {}) -> int:
    with orm.db_session:
        i = Fsm(state=state, data=data)
    return i.id


@orm.db_session
def lock(state: str) -> JSONDict:
    if state.startswith(prefix_locked):
        raise NotAllowed(f"{prefix_locked}* is not allowed: {state}")
    ts = datetime.now() - ageing_time
    i = orm.select(
        i for i in Fsm if i.ts > ts and i.state == state
    ).order_by(Fsm.ts).for_update(skip_locked=True).first()
    if not i:
        raise NotFound(state)
    prev_info = i.to_dict()
    i.state = f"{prefix_locked}{i.state}"
    i.ts = datetime.now()
    return prev_info


@orm.db_session
def transit(id: int, state: str, data_patch: JSONDict = None):
    if state.startswith(prefix_locked):
        raise NotAllowed(f"{prefix_locked}* is not allowed: {state}")
    i = Fsm.get_for_update(id=id)
    if not i:
        raise NotFound(id)
    if not i.state.startswith(prefix_locked):
        raise NotAllowed(i.state, state)
    i.state = state
    i.ts = datetime.now()
    if data_patch is not None:
        i.data.update(data_patch)


@orm.db_session
def info(id) -> JSONDict:
    try:
        return Fsm[id].to_dict()
    except orm.ObjectNotFound:
        raise NotFound(id)


@orm.db_session
def list_locked() -> List[JSONDict]:
    q = orm.select(
        (i.id, i.ts, i.state) for i in Fsm if i.state.startswith(prefix_locked)
    ).order_by(2)
    return [{"id": id, "ts": ts, "state": state} for id, ts, state in q]


@orm.db_session
def list_available(state) -> List[JSONDict]:
    ts = datetime.now() - ageing_time
    q = orm.select(
        (i.id, i.ts) for i in Fsm if i.ts > ts and i.state == state
    ).order_by(2)
    return [{"id": id, "ts": ts} for id, ts in q]


def parse_db(url: str) -> Dict[str, Union[str, int, None]]:
    """See:
    https://docs.ponyorm.org/database.html#binding-the-database-object-to-a-specific-database
    """
    from os.path import abspath
    from urllib.parse import urlparse, unquote
    u = urlparse(url)
    provider = u.scheme
    if provider == "sqlite":
        filename = unquote(u.netloc + u.path)
        if filename != ":memory:":
            filename = abspath(filename)  # $CWD/filename
        return {
            "provider": provider,
            "filename": filename,
            "create_db": True
        }
    auth, _, loc = u.netloc.rpartition("@")
    user, _, password = map(unquote, auth.partition(":"))
    host, _, port = map(unquote, loc.partition(":"))
    port = port and int(port) or None
    database = u.path.lstrip("/")
    if provider == "postgres":
        return {
            "provider": provider,
            "user": user,
            "password": password,
            "host": host,
            "port": port,
            "database": database,
        }
    elif provider == "mysql":
        return {
            "provider": provider,
            "user": user,
            "passwd": password,
            "host": host,
            "port": port,
            "db": database,
        }

    raise ValueError(url)


def _init_this():
    # orm.sql_debug(True)
    db_url = getenv("FSMHUB_DB_URL") or "sqlite://fsmhub.db"
    options = parse_db(db_url)
    db.bind(**options)
    db.generate_mapping(create_tables=True)


if __name__ == '__main__':
    db.bind('sqlite', filename=':memory:')
    db.generate_mapping(create_tables=True)
    print(parse_db("sqlite://test"))
    print(parse_db("mysql://root:password@my:3306/test"))
    print(parse_db("postgres://postgres@postgres"))
else:
    _init_this()



#!/usr/bin/env python3

from pathlib import Path
from datetime import datetime
from pony.orm import Database, PrimaryKey, Required, Optional, Json, composite_index


db = Database()


@db.on_connect(provider="sqlite")
def _home_sqliterc(_, conn):
    rc = Path.home() / ".sqliterc"
    if rc.exists():
        with rc.open() as f:
            conn.executescript("".join(i for i in f if i[0] != "."))


class Fsm(db.Entity):
    id = PrimaryKey(int, auto=True)
    state = Required(str, 80)
    ts = Required(datetime, default=datetime.now)
    data = Optional(Json)
    composite_index(ts, state)


if __name__ == '__main__':
    db.bind('sqlite', filename=':memory:')
    db.generate_mapping(create_tables=True)



#!/usr/bin/env python3

from setuptools import setup

author = "lwzm"

setup(
    name="fsmhub",
    version="4.8",
    description="Finite State Machine storage hub",
    author=author,
    author_email="{}@qq.com".format(author),
    url="https://github.com/lwzm/fsmhub",
    keywords="fsm http web store".split(),
    packages=["fsmhub"],
    scripts=["fsm-step"],
    install_requires="pony fastapi uvicorn".split(),
    classifiers=[
        "Environment :: Console",
        "Programming Language :: Python :: 3",
    ],
)

# $ ./setup.py sdist bdist_wheel
# $ twine upload dist/fsmhub-xxx
